#ifndef SETUP_H
#define SETUP_H

#define _USE_MATH_DEFINES // for C++
#include <cmath>

#include <vector>
#include <stdexcept>

#include <QString>
#include <QSettings>
#include <QMap>
#include <QMetaType>
#include <QDir>
#include <QDebug>

class Vector3D
{
public:
    double X=0;
    double Y=0;
    double Z=0;
    Vector3D():Vector3D(0, 0, 0) {}
    Vector3D(const double x, const double y, const double z): X(x), Y(y), Z(z) {}

    Vector3D operator+(const Vector3D& r) const {
        return Vector3D(this->X + r.X, this->Y + r.Y, this->Z + r.Z);
    }
    Vector3D operator-(const Vector3D& r) const {
        return Vector3D(this->X - r.X, this->Y - r.Y, this->Z - r.Z);
    }

    double operator*(const Vector3D& r) const {
        return this->X * r.X + this->Y * r.Y + this->Z * r.Z;
    }
    Vector3D operator*(const double& d) const {
        return Vector3D(this->X * d,  this->Y * d, this->Z * d);
    }
    Vector3D operator/(const double& d) const {
        return Vector3D(this->X / d,  this->Y / d, this->Z / d);
    }
};

inline Vector3D operator*(const double& d, const Vector3D& r){
    return r * d;
}

inline double getCosAngle(const Vector3D& l, const Vector3D& r) {
    return l*r / std::sqrt((l*l) * (r*r));
}

struct Channel
{
    bool enabled;
    double x;
    double y;
    std::vector<int> ASICID;
    std::vector<int> chNum;
    std::vector<double> z;
    std::vector<QString> source;
    std::vector<double> threshold;
    std::vector<double> caliCoef;
};

Q_DECLARE_METATYPE(Channel);

inline QDataStream& operator<<(QDataStream& out, const Channel& v) {
    out << QVariant(v.enabled).toString() << QVariant(v.x).toString() << QVariant(v.y).toString()
        << QVariant(v.ASICID[0]).toString() << QVariant(v.ASICID[1]).toString()
        << QVariant(v.chNum[0]).toString() << QVariant(v.chNum[1]).toString()
        << QVariant(v.z[0]).toString() << QVariant(v.z[1]).toString()
        << QVariant(v.source[0]).toString() << QVariant(v.source[1]).toString()
        << QVariant(v.threshold[0]).toString() << QVariant(v.threshold[1]).toString()
        << QVariant(v.caliCoef[0]).toString() << QVariant(v.caliCoef[1]).toString();
    return out;
}

inline QDebug operator<<(QDebug dbg, const Channel &v)
{
    QDebugStateSaver saver(dbg);
    dbg.nospace() << v.enabled << v.x << v.y <<
                     v.ASICID[0] << v.ASICID[1] <<
                     v.chNum[0] << v.chNum[1] <<
                     v.z[0] << v.z[1] <<
                     v.source[0] << v.source[1] <<
                     v.threshold[0] << v.threshold[1] <<
                     v.caliCoef[0] << v.caliCoef[1];
    return dbg;
}

inline QDataStream& operator>>(QDataStream& in, Channel& v) {
    QString s;
    in >> s;
    v.enabled = QVariant(s).toBool();
    in >> s;
    v.x = QVariant(s).toDouble();
    in >> s;
    v.y = QVariant(s).toDouble();
    for (int i=0;i<2;i++)
    {
        in >> s;
        v.ASICID.push_back(QVariant(s).toInt());
    }
    for (int i=0;i<2;i++)
    {
        in >> s;
        v.chNum.push_back(QVariant(s).toInt());
    }
    for (int i=0;i<2;i++)
    {
        in >> s;
        v.z.push_back(QVariant(s).toDouble());
    }
    for (int i=0;i<2;i++)
    {
        in >> s;
        v.source.push_back(QVariant(s).toString());
    }
    for (int i=0;i<2;i++)
    {
        in >> s;
        v.threshold.push_back(QVariant(s).toDouble());
    }
    for (int i=0;i<2;i++)
    {
        in >> s;
        v.caliCoef.push_back(QVariant(s).toDouble());
    }
    return in;
}

class Setup
{
private:
    /* data */
    QString settingPath;
public:
    Setup(const QString initFile);
    ~Setup();

    void readSetup(const QString path);
    void writeSetup(const QString path="settings.ini") const;
    // data file on disk
    QString filePath;
    QString inputFormat="CSV";
    QString runID = "0";
    QDir outputDir;
    // channels
    QList<Channel> channelSettings;

    // image space setup
    // radius of projection sphere,  mm
    double R;
    // Altitude angle, rad
    int thetaBins;
    double thetaMin;
    double thetaMax;
    // Azimuthal angle, rad
    int phiBins;
    double phiMin;
    double phiMax;
    bool imageSpacechaged=false;

    // energy spectrum
    int ergBins;
    double ergMin;
    double ergMax;

    // Rejections
    bool energyCut;
    double energyLow;
    double energyUp;

    // coincidence
    double timeWindow; // ns

    // data saving options
    bool saveImage=false;
    bool saveSpectrum=false;
    QString saveImageFormat = "ROOT";
    QString saveSpectrumFormat="ROOT";


    const double sgmE = 0.06; // 6% energy resolution
//    const Vector3D sgmpos = Vector3D(0.8 / 2, 0.43 / 2, 1.0 / 2); // mm
     const Vector3D sgmpos = Vector3D(0, 0, 0); // mm
    const int order=3;

    // refresh every 10 events
    const int chuckSize = 10;

    // max number of cones to process
    int maxN=100000;
};

class Cone
{
public:
    // interaction site 1 is the apex
    Vector3D apex;
    // site 1 - site 2 is the axis
    Vector3D axis;
    // half angle = abs (scattering angle)
    // for neutron, 0 < half angle < pi /2
    // for gamma,   0 < half angle < pi
    double cosHalfAngle;
    // initial energy
    double E0;
    // energy deposited in compoton scattering
    double Edpst;
    Cone(): Cone(Vector3D(0,0,0), Vector3D(0,0,0), 0, 0, 0) {}
    Cone(const Vector3D point, const Vector3D line, const double cosTheta, const double E0_, const double Edpst_):
       apex(point), axis(line), cosHalfAngle(cosTheta), E0(E0_), Edpst(Edpst_) {}
    Cone(const std::string& record) {
        apex = Vector3D(std::stod(record.substr(0, 8)),
                        std::stod(record.substr(8, 8)),
                        std::stod(record.substr(16, 8)));
        axis = Vector3D(std::stod(record.substr(24, 8)),
                        std::stod(record.substr(32, 8)),
                        std::stod(record.substr(40, 8)));
        cosHalfAngle = std::stod(record.substr(48, 13));
        E0 = std::stod(record.substr(61, 13));
        Edpst = std::stod(record.substr(74, 13));
    }
    Cone(const QString& record) {
        apex = Vector3D(record.mid(0, 8).toDouble(),
                        record.mid(8, 8).toDouble(),
                        record.mid(16, 8).toDouble());
        axis = Vector3D(record.mid(24, 8).toDouble(),
                        record.mid(32, 8).toDouble(),
                        record.mid(40, 8).toDouble());
        cosHalfAngle = record.mid(48, 13).toDouble();
        E0 = record.mid(61, 13).toDouble();
        Edpst = record.mid(74, 13).toDouble();
    }
};

#endif // SETUP_H
