#include "worker.h"
#include "TSystem.h"
#include <QFile>
#include <QString>
#include <QTextStream>
#include <QDataStream>
#include <QDebug>
#include <QDir>

bool Worker::processCoincidence(const std::vector<MyPulse>& pulses, Cone& newCone){
    if (pulses.size() != 2)
    {
        return false;
    }
    // energy cut
    double Etot = pulses[1].height + pulses[0].height ;
    if (config->energyCut)
    {
        if (Etot < config->energyLow || Etot > config->energyUp)
        {
            return false;
        }
    }
    if (pulses[0].height < 0.01 || pulses[1].height < 0.01)
        return false;
    double cosHalfangle = 1 - pulses[0].height * 0.511 / (pulses[1].height * Etot);
    if (std::abs(cosHalfangle)>=1)
    {
        return false;
    }
    Vector3D axis = pulses[1].pos - pulses[0].pos;
    if (axis * axis == 0){
        return false;
    }
    newCone = Cone(pulses[0].pos, axis, cosHalfangle,
                   Etot, pulses[0].height);
    return true;
}

bool processCoincidence(const std::vector<Event>& pulses, Cone& newCone)
{
    if(pulses.size() < 2)
        return false;
    // pick channels with hit = 1 && value > threshold
}

Worker::Worker(QObject *parent, const Setup* config_, PlotData* plotdata_) :
    QThread(parent),
    config(config_),
    plotdata(plotdata_)
{
//    stopped=true;
    stopped = false;
    exitted=false;

    enabledChs = EnabledChannel(config->channelSettings);

    infile.setFileName(config->filePath);
    QString outfp = config->outputDir.absolutePath() + "/cones.txt";
//    qDebug()<< outfp << '\n';
    outfile.open(outfp.toStdString(), std::ios::out);

    if(config->inputFormat == "CSV")
    {
        if (!infile.open(QIODevice::ReadOnly | QIODevice::Text))
        {
            throw std::invalid_argument("Cannot open file: " + config->filePath.toStdString());
        }
    }
    else if (config->inputFormat == "BIN")
    {
        if (!infile.open(QIODevice::ReadOnly))
        {
            throw std::invalid_argument("Cannot open file: " + config->filePath.toStdString());
        }
    }
    if (!outfile.good())
    {
        throw std::invalid_argument("Cannot open file: " + outfp.toStdString());
    }
}

void Worker::run()
{
    QDataStream in(&infile);

    while(!exitted)
    {
        if (!stopped)
        {
            // read some cones
            std::vector<Cone> cones;
            if (config->inputFormat=="BIN"){
                readBIN(cones, in);
            }

            if (exitted)
                break;
            // update image
            plotdata->updateImage(cones.cbegin(), cones.cend(), true);

            if (exitted || plotdata->counts >= config->maxN)
                break;
        }
    }
    // save image & spectrum before quitting
    qDebug() << "Save results before exiting worker thread";
    plotdata->saveOnExit();
    qDebug() << "Worker thread exited.";
}

//void Worker::readCSV(std::vector<Cone> &cones, QTextStream& in)
//{
//    QString line;
//    std::vector<ulong> histNums;
//    static std::vector<MyPulse> pulses;
//    static double prevTT=0;
//    int i=0;
//    while (!exitted && i < config->chuckSize && in.readLineInto(&line))
//    {
//        MyPulse newPulse = MyPulse(line);
//        auto it = enabledChs.indexOf(newPulse.cellNo);
//        if (it < 0){
//            // cell not found
//            // go to next line
//            continue;
//        }
//        else {
//            newPulse.pos.X = config->channelSettings[it].x;
//            newPulse.pos.Y = config->channelSettings[it].y;
//            newPulse.height *= config->channelSettings[it].caliCoef;
//            // TO DO: calculate z
//        }
//        if (pulses.size()==0 || std::abs(newPulse.timeStamp - prevTT) < config->timeWindow){
//            prevTT = newPulse.timeStamp;
//            pulses.push_back(std::move(newPulse));
//        }
//        else {
//            // process coincident events
//            Cone newCone;
//            if (processCoincidence(pulses, newCone)){
//                histNums.push_back(pulses[0].histNo);
//                cones.push_back(std::move(newCone));
//                i++;
//            }
//            pulses.clear();
//            prevTT = newPulse.timeStamp;
//            pulses.push_back(std::move(newPulse));
//        }
//    }

//    // save cones to file
//    for (int i=0;i<cones.size();i++)
//    {
//        outfile << std::fixed << std::setprecision(2)
//                << std::setw(8) << cones[i].apex.X // << "    "
//                << std::setw(8) << cones[i].apex.Y // << "    "
//                << std::setw(8) << cones[i].apex.Z // << "    "
//                << std::setw(8) << cones[i].axis.X // << "    "
//                << std::setw(8) << cones[i].axis.Y // << "    "
//                << std::setw(8) << cones[i].axis.Z // << "    "
//                << std::fixed << std::setprecision(8)
//                << std::setw(13) << cones[i].cosHalfAngle
//                << std::setw(13) << cones[i].E0
//                << std::setw(13) << cones[i].Edpst
//                << std::setw(13) << histNums[i] << '\n';
//    }

//}

void Worker::readBIN(std::vector<Cone> &cones, QDataStream &in)
{
    // hold a single event
    char buffer[eventSize];
    static std::vector<Event> pulses;
    static double prevTT=0;
    int i = 0;
    int len = 0;
    while (!exitted && i < config->chuckSize) {
        while (!exitted && i < config->chuckSize) {
            // read data into buffer
            if (len != eventSize)
            {
                int tmp=in.readRawData(buffer + len, eventSize-len);
                if(tmp < 0)
                    break;
                len += tmp;
            }
            if(len != eventSize)
                break;
            else
                len = 0;
            // parse data
            Event newPulse;
            DecodeCITIROCRowEvents(buffer, eventSize, newPulse);
            if (pulses.size()==0 || std::abs(newPulse.EventTimecode_ns - prevTT) < config->timeWindow){
                prevTT = newPulse.EventTimecode_ns;
                pulses.push_back(std::move(newPulse));
            }
            else {
                // process coincident events
                Cone newCone;
                if (processCoincidence(pulses, newCone)){
                    cones.push_back(std::move(newCone));
                    i++;
                }
                pulses.clear();
                prevTT = newPulse.EventTimecode_ns;
                pulses.push_back(std::move(newPulse));
            }
        }
        // wait for new data
        if (in.atEnd())
            msleep(1000);
    }
}

int Worker::DecodeCITIROCRowEvents(const char* buf, const int bufSize, Event& DataCITIROCA)
{
    //Convert char buffer in uin32
    const uint32_t* bufferA = (uint32_t*)&(buf[0]);
    const uint32_t valid_wordA = bufSize/4;

    int DecodedPackets = 0;
    double Time =0;
    double minTime = 0;
    uint32_t i, j, t, s;
    uint datarow[97];

    t= 0;
    s=0;
    while (t < valid_wordA)
    {
        switch (s) {
        case 0:
            if(((bufferA[t] >> 4) & 0xc000000) == 0x8000000)
            {
                s = 1;
                DataCITIROCA.AsicID = (uint16_t) (bufferA[t] & 0xF);
                DataCITIROCA.EventTimecode = ((uint64_t)bufferA[t + 1]);
                DataCITIROCA.RunEventTimecode = (((uint64_t)bufferA[t+ 2]) ) + (((uint64_t)bufferA[t + 3]) << 32);
                DataCITIROCA.EventCounter = ((uint64_t)bufferA[t + 4]);

                DataCITIROCA.EventTimecode_ns = DataCITIROCA.EventTimecode * 25;
                DataCITIROCA.RunEventTimecode_ns =DataCITIROCA.RunEventTimecode * 25;

                t = t + 5;
                minTime = 100000000000000;
            }
            else
                t++;
            break;

        case 1:
            for(i = 0; i < 32; i++)
            {
                uint32_t bword = bufferA[t];
                datarow[i * 3 + 0] = (bword >> 0) & 0x3FFF;
                datarow[i * 3 + 1] = (bword >> 14) & 0x3FFF;
                datarow[i * 3 + 2] = (bword >> 28) & 0x1;
                t++;
            }
            for(i = 0; i < 32; i++)
            {
                DataCITIROCA.hit[i] = (bool)((datarow[i*3+2] & 0x1) == 1 ? true: false);
                int dataHG = (int)datarow[(i * 3) + 0];
                int dataLG= (int)datarow[(i * 3) + 1];
//                if(dataHG > ThresholdSoftware)
//                    DataCITIROCA.chargeHG[i] = (ushort)dataHG;
//                else
//                    DataCITIROCA.chargeHG[i] = 0;
//                if(dataHG > ThresholdSoftware)
//                    DataCITIROCA.chargeLG[i] = (ushort)dataLG;
//                else
//                    DataCITIROCA.chargeLG[i] = 0;
                DataCITIROCA.chargeHG[i] = (ushort)dataHG;
                DataCITIROCA.chargeLG[i] = (ushort)dataLG;
            }
            s = 2;
            break;

        case 2:
            if((bufferA[t] & 0xc0000000) == 0xc0000000)
            {
//                pC.Enqueue(DataCITIROCA);
                DecodedPackets++;
            }
            t++;
            s = 0;
            break;
        }
    }
    return DecodedPackets;
}
void Worker::handleStart()
{
    stopped=false;
}

void Worker::handlePause()
{
    stopped=true;
}

void Worker::handleStop()
{
    exitted = true;
    qDebug() << "Exiting thread signal received.";
}
